@@ -18,7 +18,7 @@ use namespace::autoclean;
 
 extends 'App::Sqitch::Engine';
 
-our $VERSION = '0.9996';
+our $VERSION = 'v1.4.1'; # VERSION
 
 has registry_uri => (
     is       => 'ro',
@@ -59,10 +59,14 @@ sub registry_destination {
     return $uri->as_string;
 }
 
+sub _def_user { $ENV{ISC_USER} }
+sub _def_pass { $ENV{ISC_PASSWORD} }
+
 has dbh => (
     is      => 'rw',
     isa     => DBH,
     lazy    => 1,
+    clearer => '_clear_dbh',
     default => sub {
         my $self = shift;
         my $uri  = $self->registry_uri;
@@ -123,7 +127,7 @@ has tz_offset => (
     isa      => Maybe[Int],
     lazy     => 1,
     default => sub {
-        # From: http://stackoverflow.com/questions/2143528/whats-the-best-way-to-get-the-utc-offset-in-perl
+        # From: https://stackoverflow.com/questions/2143528/whats-the-best-way-to-get-the-utc-offset-in-perl
         my @t = localtime(time);
         my $gmt_offset_in_seconds = timegm(@t) - timelocal(@t);
         my $offset = -($gmt_offset_in_seconds / 3600);
@@ -180,7 +184,7 @@ sub is_deployed_tag {
     }, undef, $tag->id)->[0];
 }
 
-sub initialized {
+sub _initialized {
     my $self = shift;
 
     # Try to connect.
@@ -197,7 +201,7 @@ sub initialized {
     }, undef, 'CHANGES')->[0];
 }
 
-sub initialize {
+sub _initialize {
     my $self = shift;
     my $uri  = $self->registry_uri;
     hurl engine => __x(
@@ -237,6 +241,8 @@ sub connection_string {
         'Database name missing in URI {uri}',
         uri => $uri,
     );
+    # Use _port instead of port so it's empty if no port is in the URI.
+    # https://github.com/sqitchers/sqitch/issues/675
     my $host = $uri->host   or return $file;
     my $port = $uri->_port  or return "$host:$file";
     return "$host/$port:$file";
@@ -285,7 +291,11 @@ sub _no_table_error  {
 }
 
 sub _no_column_error  {
-    return $DBI::errstr && $DBI::errstr =~ /^-Column unknown|/m;
+    return $DBI::errstr && $DBI::errstr =~ /^-Column unknown/m;
+}
+
+sub _unique_error  {
+    return $DBI::errstr && $DBI::errstr =~ /no 2 table rows can have duplicate column values$/m;
 }
 
 sub _regex_op { 'SIMILAR TO' }               # NOT good match for
@@ -340,6 +350,13 @@ sub run_upgrade {
     my @cmd    = $self->isql;
     $cmd[-1]   = $self->connection_string($uri);
     my $sqitch = $self->sqitch;
+    unless ($uri->host) {
+        # Only one connection allowed when using an embedded database (Engine 12
+        # provider). So disconnect so that the upgrade can connect and succeed,
+        # and clear the disconnected handle so that the next call to ->dbh will
+        # reconnect.
+        $self->dbh->disconnect; $self->_clear_dbh;
+    }
     $sqitch->run( @cmd, '-input' => $sqitch->quote_shell($file) );
 }
 
@@ -608,7 +625,7 @@ sub change_offset_from_id {
         SELECT FIRST 1 $offset_expr
                c.change_id AS "id", c.change AS name, c.project, c.note,
                $tscol AS "timestamp", c.planner_name, c.planner_email,
-               $tagcol AS tags
+               $tagcol AS tags, c.script_hash
           FROM changes   c
           LEFT JOIN tags t ON c.change_id = t.change_id
          WHERE c.project = ?
@@ -616,7 +633,7 @@ sub change_offset_from_id {
                SELECT committed_at FROM changes WHERE change_id = ?
          )
          GROUP BY c.change_id, c.change, c.project, c.note, c.planned_at,
-               c.planner_name, c.planner_email, c.committed_at
+               c.planner_name, c.planner_email, c.committed_at, c.script_hash
          ORDER BY c.committed_at $dir
     }, undef, $self->plan->project, $change_id ) || return undef;
     $change->{timestamp} = _dt $change->{timestamp};
@@ -654,11 +671,10 @@ sub change_id_for {
     if ( my $change = $p{change} ) {
         if ( my $tag = $p{tag} ) {
             # There is nothing before the first tag.
-            return undef if $tag eq 'ROOT' || $tag eq 'FIRST';
+            return undef if $tag eq 'ROOT';
 
             # Find closest to the end for @HEAD.
-            return $self->_cid_head($project, $change)
-                if $tag eq 'HEAD' || $tag eq 'LAST';
+            return $self->_cid_head($project, $change) if $tag eq 'HEAD';
 
             # Find by change name and following tag.
             return $dbh->selectcol_arrayref(q{
@@ -676,7 +692,7 @@ sub change_id_for {
 
         # Find earliest by change name.
         my $ids = $dbh->selectcol_arrayref(qq{
-            SELECT FIRST 1 change_id
+            SELECT change_id
               FROM changes
              WHERE project = ?
                AND changes.change  = ?
@@ -684,17 +700,16 @@ sub change_id_for {
         }, undef, $project, $change);
 
         # Return the ID.
+        return $ids->[0] if $p{first};
         return $self->_handle_lookup_index($change, $ids);
     }
 
     if ( my $tag = $p{tag} ) {
         # Just return the latest for @HEAD.
-        return $self->_cid('DESC', 0, $project)
-            if $tag eq 'HEAD' || $tag eq 'LAST';
+        return $self->_cid('DESC', 0, $project) if $tag eq 'HEAD';
 
         # Just return the earliest for @ROOT.
-        return $self->_cid('ASC', 0, $project)
-            if $tag eq 'ROOT' || $tag eq 'FIRST';
+        return $self->_cid('ASC', 0, $project) if $tag eq 'ROOT';
 
         # Find by tag name.
         return $dbh->selectcol_arrayref(q{
@@ -801,23 +816,31 @@ sub log_deploy_change {
         planner_email
         committed_at
     ));
-    $dbh->do(qq{
-        INSERT INTO changes (
-            $cols
-        )
-        VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, $ts)
-    }, undef,
-        $id,
-        $change->script_hash,
-        $name,
-        $proj,
-        $change->note,
-        $user,
-        $email,
-        $self->_char2ts( $change->timestamp ),
-        $change->planner_name,
-        $change->planner_email,
-    );
+    try {
+        $dbh->do(qq{
+            INSERT INTO changes (
+                $cols
+            )
+            VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, $ts)
+        }, undef,
+            $id,
+            $change->script_hash,
+            $name,
+            $proj,
+            $change->note,
+            $user,
+            $email,
+            $self->_char2ts( $change->timestamp ),
+            $change->planner_name,
+            $change->planner_email,
+        );
+    } catch {
+        hurl engine => __x(
+            'Cannot log change "{change}": The deploy script is not unique',
+            change => $name,
+        ) if $self->_unique_error;
+        die $_;
+    };
 
     if ( my @deps = $change->dependencies ) {
         foreach my $dep (@deps) {
@@ -865,8 +888,8 @@ sub log_deploy_change {
 }
 
 sub default_client {
-    my $self   = shift;
-    my $ext    = $^O eq 'MSWin32' || $^O eq 'cygwin' ? '.exe' : '';
+    my $self = shift;
+    my $ext  = App::Sqitch::ISWIN || $^O eq 'cygwin' ? '.exe' : '';
 
     # Create a script to run.
     require File::Temp;
@@ -889,7 +912,8 @@ sub default_client {
         my $loops = 0;
         for my $dir (File::Spec->path) {
             my $path = file $dir, $try;
-            $path = Win32::GetShortPathName($path) if $^O eq 'MSWin32';
+            # GetShortPathName returns undef for nonexistent files.
+            $path = Win32::GetShortPathName($path) // next if App::Sqitch::ISWIN;
             if (-f $path && -x $path) {
                 if (try { App::Sqitch->probe($path, @opts) =~ /Firebird/ } ) {
                     # Restore STDERR and return.
@@ -972,7 +996,7 @@ David E. Wheeler <david@justatheory.com>
 
 =head1 License
 
-Copyright (c) 2012-2015 iovation Inc.
+Copyright (c) 2012-2024 iovation Inc., David E. Wheeler
 
 Copyright (c) 2013 Ștefan Suciu
 
